/* $Id: cube.c,v 1.3 2003/01/28 09:11:15 d Exp $ */
/* David Leonard, 2003. Public domain. */

/*
 * Rubik's cube simulation
 */

#include <stdio.h>
#include <unistd.h>
#include "cube.h"

/*#define DEBUG*/
/*#define MAIN */

/*
 * A rubik's cube has 6 faces, and 9 cells per face, with 6 different colours.
 */
#define MAX_FACE 6

typedef unsigned char colour_t;		/* '0'..'5' */
#define MAX_COLOUR MAX_FACE

#define MAX_SQUARES (MAX_FACE * 9)

typedef colour_t state_t[MAX_SQUARES];
#define GOAL_STATE \
  { '0','0','0','0','0','0','0','0','0', \
    '1','1','1','1','1','1','1','1','1', \
    '2','2','2','2','2','2','2','2','2', \
    '3','3','3','3','3','3','3','3','3', \
    '4','4','4','4','4','4','4','4','4', \
    '5','5','5','5','5','5','5','5','5' }

#define _SQUARE_COLOUR(s, face, row, col)  ((s)[(face)*9 + (row)*3 + (col)])


#ifdef DEBUG
#define SQUARE_COLOUR	square_colour
colour_t
square_colour(state_t s, int face, int row, int col)
{
	colour_t c;

	if (face < 0 || face >=6) {
		fprintf(stderr, "SQUARE_COLOUR: bad face %d\n", face);
		return '0';
	}
	if (row < 0 || row >=3 || col < 0 || col >= 3) {
		fprintf(stderr, "SQUARE_COLOUR: bad row,col %d,%d\n", 
			row, col);
		return '0';
	}

	c = _SQUARE_COLOUR(s, face, row, col);
	if (c < '0' || c > '5') {
		fprintf(stderr, "bad colour 0x%02x at face=%d row=%d col=%d\n",
			c, face, row, col);
		return '0';
	}
	return c;
}
#else
#define SQUARE_COLOUR(s,f,r,c)	_SQUARE_COLOUR(s,f,r,c)
#endif

/*
 * There are 3 types of rotation: about the x, y and z axes respectively.
 * Each rotation can be at one of 3 depths. (There is also a choice of
 * one of two directions but rotating three times in one direction is equivalent
 * to one rotation in the other, so we simplify to only one direction)
 * 
 * Actually, because rotations at depth 1 are equivalent to two consecutive
 * reverse rotations at depths 0 and 2, there are really only 6 distinct
 * moves to make on the cube.
 */
typedef unsigned char move_t;		/* 0..8 */
#define MAX_MOVE 9

struct node {
	state_t state;
	struct node* move[MAX_MOVE];
};

static void init_transform(void);
static void move(state_t in, move_t move, state_t out);
static void print(state_t s);

static unsigned char transform[MAX_MOVE][MAX_SQUARES];

static void
init_transform()
{
    static char *chain[3][2] = {
        /* move forward by 3 */		      /* move forward by 2 */
     { "28 25 22 12 11 10 56 57 58 30 33 36", "00 01 02 05 08 07 06 03" },
     { "27 24 21 15 14 13 53 54 55 31 34 37", NULL },
     { "26 23 20 18 17 16 50 51 52 32 35 38", "40 43 46 47 48 45 42 41" }
    };
    int move, i, j;

    for (move = 0; move < MAX_MOVE; move++) {
	int face = move / 3;
	int depth = move % 3;
	int facemap[6];

	for (i = 0; i < MAX_SQUARES; i++)
	    transform[move][i] = i;

	facemap[0] = face;
	facemap[1] = (face + 1) % 3;
	facemap[2] = (face + 2) % 3;
	facemap[3] = 3 + (3 - face) % 3;
	facemap[4] = 3 + (4 - face) % 3;
	facemap[5] = 5 - face;

	for (j = 0; j < 12; j++) {
		int j2 = (j + 3) % 12;
		char *src = chain[depth][0] + j*3;
		char *dst = chain[depth][0] + j2*3;
		int s = 9 * facemap[src[0]-'0'] + src[1]-'0';
		int d = 9 * facemap[dst[0]-'0'] + dst[1]-'0';
		transform[move][d] = s;
	}

	if (chain[depth][1]) 
	  for (j = 0; j < 8; j++) {
		int j2 = (j + 2) % 8;
		char *src = chain[depth][1] + j*3;
		char *dst = chain[depth][1] + j2*3;
		int s = 9 * facemap[src[0]-'0'] + src[1]-'0';
		int d = 9 * facemap[dst[0]-'0'] + dst[1]-'0';
		transform[move][d] = s;
	  }
    }
}

#ifdef DEBUG
static void
printface(state_t s, int f, int r, int c)
{
	/*
	 * Rubik's original cube had these opposing colours:
	 *  blue-green
	 *  red-orange
	 *  yellow-white
	 */
#define COL_RED		"41;37"
#define COL_GREEN	"42"
#define COL_YELLOW	"43"
#define COL_BLUE	"44;37"
#define COL_PURPLE	"45;37"
#define COL_CYAN	"46"
#define COL_GREY	"47"
	static const char * col[6] = {
		COL_BLUE,
		COL_RED,
		COL_YELLOW,
		COL_GREEN,
		COL_PURPLE,
		COL_GREY
	};

	colour_t thecol = SQUARE_COLOUR(s, f, r, c);
	printf("\033[%sm%c\033[0m", col[thecol-'0'],
		r==1 && c==1 ? '0'+f : 'a'+r*3+c);
}

void
print(state_t s)
{
	int row, col;

	for (row = 0; row < 3; row++) {
	    printf("       ");
	    for (col = 0; col < 3; col++) {
		putchar(' ');
		printface(s, 2, 2-col, row);
	    }
	    printf("|\n");
	}
	printf("        _ _ _\n");
	for (row = 0; row < 3; row++) {
	    printf(" ");
	    for (col = 0; col < 3; col++) {
	      printface(s, 3, 2-row, 2-col);
	      putchar(' ');
	    }

	    putchar(' ');
	    for (col = 0; col < 3; col++) {
	      printface(s, 0, row, col);
	      putchar(' ');
	    }

	    putchar('|');
	    for (col = 0; col < 3; col++) {
	      printface(s, 1, col, 2-row);
	      putchar(' ');
	    }

	    for (col = 0; col < 3; col++) {
	      putchar(' ');
	      printface(s, 4, 2-col, row);
	    }

	    printf("|\n");
	}
	printf(" ~ ~ ~\n");

	for (row = 0; row < 3; row++) {
	    printf("        ");
	    for (col = 0; col < 3; col++) {
	        printface(s, 5, 2-row, 2-col);
		putchar(' ');
	    }
	    printf("\n");
	}
	printf("        ~ ~ ~\n");
}
#endif

void
move(state_t in, move_t move, state_t out)
{
	int dst, src;

	for (dst = 0; dst < MAX_SQUARES; dst++) {
		src = transform[move][dst];
		out[dst] = in[src];
	}
}

#ifdef MAIN
int
main()
{
	char cmd[10], *c;
	state_t s = GOAL_STATE;
	state_t t;
	int moveno = 0;
	int tty = isatty(STDIN_FILENO);

	/* srand(time(0)); */

	init_transform();

	for (;;) {
		print(s);
		moveno++;
	    prompt:
		printf("%d> ", moveno);
		fflush(stdout);
		if (fgets(cmd, sizeof cmd, stdin) == NULL)
			break;
		if (!tty) printf("%s", cmd);
		c = cmd;
		if (cmd[0] == '-')
			c++;
		if (c[0] < 'x' || c[0] > 'z' ||
		    c[1] < '0' || c[1] > '2' || 
		    c[2] != '\n') {
			printf("expected -?[xyz][012]\n");
			goto prompt;
		}
		move(s, (c[0]-'x')*3 + (c[1]-'0'), t);
		if (cmd[0] != '-') {
		    move(t, (c[0]-'x')*3 + (c[1]-'0'), s);
		    move(s, (c[0]-'x')*3 + (c[1]-'0'), t);
		}
		memcpy(s, t, sizeof s);
	}

	exit(0);
}
#endif

static state_t thestate;
static state_t savedstate;
void
cube_reset()
{
	static state_t goalstate = GOAL_STATE;
	static int _initialized = 0;
	if (!_initialized) {
		init_transform();
		_initialized++;
	}
	memcpy(&thestate, &goalstate, sizeof thestate);
#ifdef DEBUG
	printf("cube_reset():\n");
	print(thestate);
#endif
}

void
cube_turn(int face, int clockwise)
{
	static const int opmap[6] = { 0, 2, 8, 6, 3, 5 };
	static const int cwmap[6] = { 0, 1, 1, 0, 0, 1 };
	state_t tmp;
	int op;

#ifdef DEBUG
	if (face < 0 || face >= 6) {
		fprintf(stderr, "cube_turn: bad face %d\n", face);
		return;
	}
#endif
	op = opmap[face];
	move(thestate, op, tmp);
	if (clockwise ? cwmap[face] : !cwmap[face]) {
		move(tmp, op, thestate);
		move(thestate, op, tmp);
		move(tmp, op, thestate);
	}
	memcpy(thestate, tmp, sizeof thestate);
#ifdef DEBUG
	printf("cube_turn(face=%d clockwise=%d) ->\n", face, clockwise);
	print(thestate);
#endif
}

void
cube_roll(int face, int clockwise)
{
	static const int opmap[3] = { 0, 6, 3 };
	static const int cwmap[3] = { 1, 0, 1 };
	state_t tmp;
	int i, ff;

	if (face % 2 == 1) {
		face = face - 1;
		clockwise = !clockwise;
	}
	face /= 2;

#ifdef DEBUG
	if (face < 0 || face > 2)
		fprintf(stderr, "cube_roll: bad face/2 %d\n", face);
#endif

	if (cwmap[face])
		clockwise = !clockwise;
	ff = opmap[face];
	
	move(thestate, ff, tmp);
	move(tmp, ff+1, thestate);
	move(thestate, ff+2, tmp);
	if (clockwise) {
		move(tmp, ff, thestate);
		move(thestate, ff+1, tmp);
		move(tmp, ff+2, thestate);
		move(thestate, ff, tmp);
		move(tmp, ff+1, thestate);
		move(thestate, ff+2, tmp);
	}
	memcpy(thestate, tmp, sizeof thestate);
}

int
cube_get(int face, int x, int y)
{
	return SQUARE_COLOUR(thestate, face, y, x) - '0';
}

void
cube_checkpoint()
{
	memcpy(savedstate, thestate, sizeof savedstate);
}

int
cube_changed(int face, int x, int y)
{
	return	SQUARE_COLOUR(savedstate,face,y,x) != 
		SQUARE_COLOUR(thestate,face,y,x);
}
